@@ -13,13 +13,13 @@
 use quote::quote;
 use std::collections::HashSet;
 use std::fs;
+use std::path::PathBuf;
 use std::process::Command;
 use std::str::FromStr;
 use syn::parse::{Parse, ParseStream};
 use syn::punctuated::Punctuated;
 use syn::visit::{self, Visit};
 use syn::{AttrStyle, File, Ident, ItemImpl, ItemStruct, Path, Token, Type, TypePath};
-
 /// Visitor scanning rust source code for struct belonging to enums
 struct Visitor {
     // Identifiers of structs belonging to Operation enum
@@ -34,6 +34,8 @@ struct Visitor {
     pragma_operations: Vec<Ident>,
     // Identifiers of structs belonging to PragmaNoiseOperation enum
     pragma_noise_operations: Vec<Ident>,
+    // Identifiers of structs belonging to PragmaNoiseProbaOperation enum
+    pragma_noise_proba_operations: Vec<Ident>,
     // Identifiers of structs belonging to GateOperation enum
     gate_operations: Vec<Ident>,
     // Identifiers of structs belonging to Rotation enum
@@ -59,6 +61,7 @@ impl Visitor {
             multi_qubit_operations: Vec::new(),
             pragma_operations: Vec::new(),
             pragma_noise_operations: Vec::new(),
+            pragma_noise_proba_operations: Vec::new(),
             gate_operations: Vec::new(),
             rotations: Vec::new(),
             definitions: Vec::new(),
@@ -145,6 +148,13 @@ impl<'ast> Visit<'ast> for Visitor {
                 {
                     self.pragma_noise_operations.push(i.ident.clone());
                 }
+                if parsed_arguments.contains("Operate")
+                    && parsed_arguments.contains("OperatePragma")
+                    && parsed_arguments.contains("OperatePragmaNoise")
+                    && parsed_arguments.contains("OperatePragmaNoiseProba")
+                {
+                    self.pragma_noise_proba_operations.push(i.ident.clone());
+                }
                 if parsed_arguments.contains("Operate") && parsed_arguments.contains("OperateGate")
                 {
                     self.gate_operations.push(i.ident.clone());
@@ -219,6 +229,9 @@ impl<'ast> Visit<'ast> for Visitor {
                 if trait_name.as_str() == "OperatePragmaNoise" {
                     self.pragma_noise_operations.push(id.clone());
                 }
+                if trait_name.as_str() == "OperatePragmaNoiseProba" {
+                    self.pragma_noise_proba_operations.push(id.clone());
+                }
                 if trait_name.as_str() == "OperateMultiQubitGate" {
                     self.two_qubit_gate_operations.push(id);
                 }
@@ -286,13 +299,21 @@ fn main() {
         #v(#v)}
     });
     // Construct TokenStreams for variants of pragma enum
-    let pragma_noise_operations_quotes = vis.pragma_noise_operations.into_iter().map(|v| {
+    let pragma_noise_operations_quotes = vis.pragma_noise_operations.clone().into_iter().map(|v| {
         let msg = format!("Variant for {}", v);
         quote! {
         #[doc = #msg]
         #v(#v)}
     });
     // Construct TokenStreams for variants of pragma enum
+    let pragma_noise_proba_operations_quotes =
+        vis.pragma_noise_proba_operations.into_iter().map(|v| {
+            let msg = format!("Variant for {}", v);
+            quote! {
+            #[doc = #msg]
+            #v(#v)}
+        });
+    // Construct TokenStreams for variants of pragma enum
     let gate_operations_quotes = vis.gate_operations.into_iter().map(|v| {
         let msg = format!("Variant for {}", v);
         quote! {
@@ -348,7 +369,7 @@ fn main() {
     // Construct TokenStream for auto-generated rust file containing the enums
     let final_quote = quote! {
 
-        use crate::operations::*;
+        //use crate::operations::*;
 
         /// Enum of all Operations implementing [Operate]
         #[derive(Debug, Clone, PartialEq, InvolveQubits, Operate, Substitute)]
@@ -392,6 +413,13 @@ fn main() {
             #(#pragma_noise_operations_quotes),*
         }
 
+        /// Enum of all Operations implementing [OperatePragmaNoiseProba]
+        #[derive(Debug, Clone, PartialEq, InvolveQubits, Operate, OperateTryFromEnum, Substitute, OperatePragma, OperatePragmaNoise, OperatePragmaNoiseProba)]
+        #[cfg_attr(feature = "serialize", derive(serde::Serialize, serde::Deserialize))]
+        pub enum PragmaNoiseProbaOperation {
+            #(#pragma_noise_proba_operations_quotes),*
+        }
+
         /// Enum of all Operations implementing [OperateGate]
         #[derive(Debug, Clone, PartialEq, InvolveQubits, Operate, OperateTryFromEnum, Substitute, OperateGate)]
         #[cfg_attr(feature = "serialize", derive(serde::Serialize, serde::Deserialize))]
@@ -444,13 +472,12 @@ fn main() {
 
     };
     let final_str = format!("{}", final_quote);
-    if std::env::var("DOCS_RS").is_err() {
-        // Write to file
-        fs::write("src/operations/_auto_generated_operations.rs", final_str)
-            .expect("Could not write to file");
-        // Try to format auto generated operations
-        let _unused_output = Command::new("rustfmt")
-            .arg("src/operations/_auto_generated_operations.rs")
-            .output();
-    }
+    let out_dir = PathBuf::from(
+        std::env::var("OUT_DIR").expect("Cannot find a valid output directory for code generation"),
+    )
+    .join("_auto_generated_operations.rs");
+    // Write to file
+    fs::write(&out_dir, final_str).expect("Could not write to file");
+    // Try to format auto generated operations
+    let _unused_output = Command::new("rustfmt").arg(&out_dir).output();
 }
